/******************************************************************************
    (c) 1998      P.J. Caulfield          patrick@pandh.demon.co.uk
                  E.M. Serrat             emserrat@geocities.com
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 ******************************************************************************
 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netdnet/dn.h>
#include <netdnet/dnetdb.h>
#include <unistd.h>


struct	sockaddr_dn		sockaddr;
struct	accessdata_dn		accessdata;
static  struct	nodeent        *binadr;
static	char			dirname[80];
static	unsigned char		buf[32760];
static	int			buflen,bufptr;
static  char                   *lasterror;

/* DECnet phase IV limits */
#define MAX_NODE      6
#define MAX_USER     12
#define MAX_PASSWORD 12
#define MAX_ACCOUNT  12
#ifndef FALSE
#define TRUE 1
#define FALSE 0
#endif

#define bool int

static int parse(char *);

/*-------------------------------------------------------------------------*/
static void usage(FILE *f, bool dnprint)
{
    if (dnprint)
    {
	fprintf(f,"\nUSAGE: dnprint [OPTIONS] 'node\"user password\"::filespec'\n\n");
    }
    else
    {
	fprintf(f, "\nUSAGE: dnsubmit [OPTIONS] 'node\"user password\"::filespec'\n\n");
    }
    fprintf(f,"NOTE: The VMS filename really should be in single quotes to\n");
    fprintf(f,"      protect it from the shell\n");

    fprintf(f,"\nOptions:\n");
    fprintf(f,"  -? -h        display this help message\n");
    fprintf(f,"  -V           show version number\n");
			    
    fprintf(f,"\nExample:\n\n");
    if (dnprint)
    {
	fprintf(f," dnprint  'serv1\"user password\"::helloworld.lis'\n");
    }
    else
    {
	fprintf(f," dnsubmit  'serv1\"user password\"::myjob.com'\n");
    }
    fprintf(f,"\n");
}
/*-------------------------------------------------------------------------*/
static int dap_setup_link(void)
{
    int sockfd;
    
    if ((sockfd=socket(AF_DECnet,SOCK_SEQPACKET,DNPROTO_NSP)) == -1) 
    {
	perror("socket");
	exit(-1);
    }

    
    if (setsockopt(sockfd,DNPROTO_NSP,SO_CONACCESS,&accessdata,
		   sizeof(accessdata)) < 0) 
    {
	perror("setsockopt");
	exit(-1);
    }
    
    sockaddr.sdn_family = AF_DECnet;
    sockaddr.sdn_flags	= 0x00;
    sockaddr.sdn_objnum	= 0x11;			/* FAL		*/
    sockaddr.sdn_objnamel	= 0x00;
    memcpy(sockaddr.sdn_add.a_addr, binadr->n_addr,6);

    if (connect(sockfd, (struct sockaddr *)&sockaddr, 
		sizeof(sockaddr)) < 0) 
    {
	perror("socket");
	exit(-1);
    }
    return sockfd;
}
/*-------------------------------------------------------------------------*/
static void dap_close_link(int sockfd)
{
    if (close(sockfd) < 0)
	perror("close");
}
/*-------------------------------------------------------------------------*/
static void	dap_exchg_config(int sockfd)
{
    int		er;
    unsigned char	buf[100];
    unsigned char	confmsg[17] = {
	0x01,	/* Message type 1 = Conf */
	0x00,	/* Flag Field		 */
	0x2E,0x04, /* BufSiz 		 */
	0x07,	/* OS type = VMS	 */
	0x03,   /* RMS-32		 */
	7,0,0,0,0, /* Version 'stuff'     */
	0xA2,	/* Sequential Org Support*/
	0x80,
	0xB0,   /* Message Blocking. OK  */
	0xFE,   /* Supp Extended Attrib  */
	0xA8,   /* Supp. Seq Access/Del  */
	0x7C	/* WildCrds/Ren/Seg Msgs */
    };
    
    if ((er=write(sockfd,confmsg,sizeof(confmsg))) < 0)
    {
	perror("Send DAP config msg");
	exit(-1);
    }			      
    if ((er=read(sockfd,buf,sizeof(buf))) < 0)
    {
	perror("Recv DAP config msg");
	exit(-1);
    }			      
    /* Assume that Peer Supports the minimal characteristis we
       setup here
    */
    if (buf[0] == 0x09)
    {
	printf("Local DAP configuration rejected by remote node\n");
	exit(-1);
    }
}
/*-------------------------------------------------------------------------*/
static	void	dap_submit(int sockfd, bool print)
{
    unsigned	char	part1[]   = {0x02,	/* ATTRIB message  */
				     0x02,	/* Lenth present   */
				     0x01,      /* length          */
				     0x00, 	/* No att field    */
				     0x03,      /* ACCESS message  */
				     0x00,	/* FLAGS	   */
				     0x08,	/* Submit          */
				     0x01 	/* I/O err is fatal*/
    }; 

    int	i,er;
    int ptr = 0;
    
    if (print) part1[6] = 1; // OPEN, the ACCOMP will print it.

    for (i = 0; i < sizeof(part1); i++) buf[ptr++] = part1[i];
    buf[ptr++] = strlen(dirname);
    for (i = 0; i < strlen(dirname); i ++) buf[ptr++] = dirname[i];
    
    if ((er=write(sockfd,buf,ptr) < 0))
    {
	perror("Send DAP Access");
	exit(-1);
    }
    buflen = bufptr = 0;

}
/*-------------------------------------------------------------------------*/
static int dap_read_reply(int sockfd)
{
    
/* Read message and check status */
    buflen = read(sockfd, buf, sizeof(buf));
    if (buflen <= 0)
    {
	perror("Receiving DAP reply");
	exit(-1);
    }

    bufptr = 1;
    switch (buf[0])
    {
    case 0x02: // ATTRIB
	break;
    case 0x06: // ACK
	break;
    case 0x09: // STATUS
	if (buf[bufptr]&2) bufptr++;
	bufptr++;
	if (buf[bufptr] == 062) return -2; // File not found
	return -1;
	break;
    case 0x07: //ACCOMP
	break;
    case 0x0f: //NAME
	break;
    default:
	printf("Unknown mesage received: 0x%x\n", buf[0]);
	return -1;
	break;
    }
    return 0;
}
/*-------------------------------------------------------------------------*/
int dap_send_accomp(int sockfd)
{
// ACCOMP message 0x04 is END-OF-STREAM -- CHECK THIS PJC
    unsigned char	accomp[] = {0x07, 0x00, 0x01,  // 0x01 is CLOSE
				    0x80, 0x80, 0x40 };
    int	er;

    if ((er=write(sockfd,accomp,sizeof(accomp))) < 0)
    {
        printf("Send DAP access complete msg failed: %s", strerror(errno));
	return -1;
    }
    return 0;
}

/*-------------------------------------------------------------------------*/
static void	adjust_bufptr(int sockfd)
{
    if (buflen <= bufptr)
    {
	buflen=read(sockfd,buf,sizeof(buf));
	if (buflen < 0) 
	{
	    perror("DAP Read Directory Entry");
	    exit(-1);
	}
	bufptr=0;
    }
}
/*-------------------------------------------------------------------------*/
int main(int argc, char *argv[])
{
    int	    opt,retval;
    char    name[256];
    int     interactive = 0;
    int     verbose = 0;
    int     two_links = 0;
    int     dir_sock, del_sock;
    bool    dnprint = FALSE;

    /*	Zero Structure	AccessData			*/
    memset(&accessdata,0,sizeof(accessdata));

    // Work out the command name
    if (strstr(argv[0], "dnprint"))
    {
	dnprint = TRUE;
    }
    
    if (argc < 2)
    {
	usage(stderr, dnprint);
	exit(0);
    }

/* Get command-line options */
    opterr = 0;
    optind = 0;
    while ((opt=getopt(argc,argv,"?hV")) != EOF)
    {
	switch(opt)
	{
	case 'h': 
	    usage(stdout, dnprint);
	    exit(1);
	    
	case '?':
	    usage(stderr, dnprint);
	    exit(1);

	case 'V':
	    printf("\ndnsubmit from dntools version %s\n\n", VERSION);
	    exit(1);
	    break;
	}
    }

    /* Parse the filename into its parts */
    if (!parse(argv[optind]))
    {
	fprintf(stderr, "%s\n", lasterror);
	exit(2);
    }
    
    dir_sock = dap_setup_link(); 
    dap_exchg_config(dir_sock);

    dap_submit(dir_sock, dnprint);
    retval = dap_read_reply(dir_sock);
    if (retval < 0)
    {
	printf("Error accessing %s: ", dirname);
	dnprint=FALSE; // Don't try the rest
    }

    if (dnprint)
    {
	dap_send_accomp(dir_sock);
	retval = dap_read_reply(dir_sock);
	if (retval < 0) printf("Error printing %s: ", dirname);
    }
    if (retval == -2) printf("File not found");
    if (retval < 0)   printf("\n");
    
    dap_close_link(dir_sock);
    return 0;
}
/*-------------------------------------------------------------------------*/
/* Parse the filename into its component parts */
static int parse(char *fname)
{
    enum  {NODE, USER, PASSWORD, ACCOUNT, NAME, FINISHED} state;
    int   n0=0;
    int   n1=0;
    char  node[MAX_NODE+1];
    char *nodenum;
    char *local_user;
    char  lastchar;
    
    state = NODE; /* Node is mandatory */

    while (state != FINISHED)
    {
	switch (state)
	{
	case NODE:
	    if (fname[n0] != ':' && fname[n0] != '\"' && fname[n0] != '\'')
	    {
		if (n1 >= MAX_NODE || 
		    fname[n0] == ' ' || fname[n0] == '\n')
		{
		    lasterror = "File name parse error";
		    return FALSE;
		}
		node[n1++] = fname[n0++];
	    }
	    else
	    {
		node[n1] = '\0';
		n1 = 0;
		if (fname[n0] == '\"')
		{
		    n0++;
		    state = USER;
		}
		else
		{
		    n0 += 2;
		    state = NAME;
		}
	    }
	    break;

	case USER:
	    if (fname[n0] != ' ' && fname[n0] != '\"' && fname[n0] != '\'')
	    {
		if (n1 >= MAX_USER)
		{
		    lasterror = "File name parse error";
		    return FALSE;
		}
		accessdata.acc_user[n1++] = fname[n0++];
	    }
	    else
	    {
		accessdata.acc_user[n1] = '\0';
		n1 = 0;
		if (fname[n0] == ' ')
		{
		    state = PASSWORD;
		    n0++;
		}
		else /* Must be a quote */
		{
		    state = NAME;
		    /* Check for :: */
		    n0 += 3;
		}
	    }
	    break;

	case PASSWORD:
	    if (fname[n0] != ' ' && fname[n0] != '\"' && fname[n0] != '\'')
	    {
		if (n1 >= MAX_PASSWORD)
		{
		    lasterror = "File name parse error";
		    return FALSE;
		}
		accessdata.acc_pass[n1++] = fname[n0++];
	    }
	    else
	    {
		accessdata.acc_pass[n1] = '\0';
		n1 = 0;
		if (fname[n0] == ' ')
		{
		    n0++;
		    state = ACCOUNT;
		}
		else /* Must be a quote */
		{
		    state = NAME;
		    /* Check for :: */
		    n0 += 3;
		}
	    }
	    break;

	case ACCOUNT:
	    if (fname[n0] != '\'' && fname[n0] != '\"')
	    {
		if (n1 >= MAX_ACCOUNT)
		{
		    lasterror = "File name parse error";
		    return FALSE;
		}
		accessdata.acc_acc[n1++] = fname[n0++];
	    }
	    else
	    {
		accessdata.acc_acc[n1] = '\0';;
		state = NAME;
		n1 = 0;
		/* Check for :: */
		n0 += 3;
	    }
	    break;

	case NAME:
	    strcpy(dirname, fname+n0);
	    state = FINISHED;
	    break;

	case FINISHED: // To keep the compiler happy
	    break;
	} /* switch */
    }

/* tail end validation */
    binadr = getnodebyname(node);
    if (!binadr)
    {
	lasterror = "Unknown or invalid node name ";
	return FALSE;
    }

    lastchar = dirname[strlen(dirname)-1];
    if (lastchar == ']' || lastchar == ':' || dirname[0] == '\0')
	fprintf(stderr, "No filename given\n");

    // Try very hard to get the local username
    local_user = getlogin();
    if (!local_user || local_user == (char *)0xffffffff)
      local_user = getenv("LOGNAME");
    if (!local_user) local_user = getenv("USER");
    if (local_user)
    {
        int i;

	strcpy((char *)accessdata.acc_acc, local_user);
	accessdata.acc_accl = strlen((char *)accessdata.acc_acc);
        for (i=0; i<accessdata.acc_accl; i++)
            accessdata.acc_acc[i] = toupper(accessdata.acc_acc[i]);
    }
    else
        accessdata.acc_acc[0] = '\0';


    /* Complete the accessdata structure */
    accessdata.acc_userl = strlen(accessdata.acc_user);
    accessdata.acc_passl = strlen(accessdata.acc_pass);
    accessdata.acc_accl  = strlen(accessdata.acc_acc);
    return TRUE;
}

